#ifndef AUDIO_H_INCL
#define AUDIO_H_INCL

#include <cstddef>
#include <fstream>
#include <list>
#include "gfx.h"
#include "liboal.h"
#include "utils.h"

#if MANIFEST_FEATURE_AUDIO || defined(AUDIO_H_FORCE_INCLUDE)

namespace Audio
{
	ALCdevice *DevHandle();
	ALCcontext *ConHandle();

	void ForceErrorCheck();

	inline void ListenerPos(const M::Vec<float, 3> &v)
	{
		alListenerfv(AL_POSITION, (float *) &v);
	}

	inline void ListenerVelocity(const M::Vec<float, 3> &v)
	{
		alListenerfv(AL_VELOCITY, (float *) &v);
	}

	// +X is forward, +Z is up
	inline void ListenerRot(const M::Quaternion &q)
	{
		q.PrepareMassTransform();
		M::Vec<float, 3> arr[2];
		arr[0] = q.MassTransform({1, 0, 0});
		arr[1] = q.MassTransform({0, 0, 1});
		alListenerfv(AL_ORIENTATION, (float *) arr);
	}

	// VECTORS MUST BE ORTHOGONAL, but not necessary normalized. Vectors are IN LOCAL SPACE.
	inline void ListenerRot(const M::Vec<float, 3> &forw, const M::Vec<float, 3> &up)
	{
		float arr[6] = {forw.x, forw.y, forw.z, up.x, up.y, up.z};
		alListenerfv(AL_ORIENTATION, arr);
	}

	class Buffer
	{
		ALuint buf;
	  public:

		Buffer(const char *fname, bool expect_stereo = 0);
		Buffer(const void *pcmdata, unsigned int datasize, unsigned int freq, bool bits16, bool stereo);
		Buffer(const Buffer &) = delete;
		Buffer(Buffer &&) = delete;
		Buffer &operator=(const Buffer &) = delete;
		Buffer &operator=(Buffer &&) = delete;
		operator ALuint() const {return buf;}

		~Buffer()
		{
			alDeleteBuffers(1, &buf);
		}
	};


	unsigned int ActiveSrcs();
	unsigned int InactiveSrcs();


	struct SrcInternal;

	class Source
	{
		SrcInternal *h;
	  public:

	  	enum class RepeatMode
	  	{
	  		single_no_keep,
	  		single_keep,
	  		loop,
	  	};
		Source(const Buffer &buf, RepeatMode mode = RepeatMode::single_keep, float volume = 1.0, float pitch = 1.0, const M::Vec<float, 3> &pos = {0,0,0}, const M::Vec<float, 3> &vel = {0,0,0}, bool rel_location = 0);
		Source(const Source &) = delete;
		Source(Source &&) = delete;
		Source &operator=(const Source &) = delete;
		Source &operator=(Source &&) = delete;
		~Source();

		void Play();
		void Stop();

		void Mode(RepeatMode n);

		void Loop(bool n);
		void Volume(float n);
		void Pitch(float n);
		void Pos(const M::Vec<float, 3> &n);
		void Vel(const M::Vec<float, 3> &n);
		void RelativePos(bool n);
		void ApplyChanges(); // DON'T FORGET TO CALL THIS
	};
}

#endif

#endif

#if !defined(AUDIO_H_INCL_MAINT) && defined(AUDIO_H_MAINT_ACCESS)
#define AUDIO_H_INCL_MAINT
#undef AUDIO_H_MAINT_ACCESS
namespace Audio
{
	#if MANIFEST_FEATURE_AUDIO
	void Init();
	void Tick();
	#else
	inline void Init() {}
	inline void Tick() {}
	#endif
}
#endif