#define GFX_H_FORCE_INCLUDE
#include <cmath>
#include <fstream>
#include <initializer_list>
#include <iostream>
#include <list>
#include <type_traits>
#include "gfx.h"
#include "libsdl.h"
#include "system.h"
#include "main.h"



#if MANIFEST_WINDOW_INITIAL_STATE == 1 && MANIFEST_WINDOW_RESIZABLE == 0
#error Fixed size window can`t be maximized. (MANIFEST_WINDOW_INITIAL_STATE == 1 && MANIFEST_WINDOW_RESIZABLE == 0)
#endif

namespace GFX
{
	static GLenum bufclear_flags = GL_COLOR_BUFFER_BIT;

	static SDL_GLContext conhandle;

	void ForceErrorCheck()
	{
		static std::string glerr;
		static bool isglerr;
		while (GLenum err = glGetError())
		{
			isglerr = 1;
			switch (err)
			{
			  case GL_INVALID_ENUM:                  glerr += "GL: Invalid enum.\n";                  break;
			  case GL_INVALID_VALUE:                 glerr += "GL: Invalid value.\n";                 break;
			  case GL_INVALID_OPERATION:             glerr += "GL: Invalid operation.\n";             break;
			  case GL_INVALID_FRAMEBUFFER_OPERATION: glerr += "GL: Invalid framebuffer operation.\n"; break;
			  case GL_OUT_OF_MEMORY:                 glerr += "GL: Out of memory.\n";                 break;
			  #ifdef GL_STACK_UNDERFLOW
			  case GL_STACK_UNDERFLOW:               glerr += "GL: Stack underflow.\n";               break;
			  #endif
			  #ifdef GL_STACK_OVERFLOW
			  case GL_STACK_OVERFLOW:                glerr += "GL: Stack overflow.\n";                break;
			  #endif
			  default:                               glerr += "GL: Unknown error.\n";                 break;
			}
		}
		if (isglerr)
			Main::Exit(glerr.c_str());
	}

    SDL_GLContext ConHandle()
    {
    	return conhandle;
    }

	Shader::Shader(const char *vsrc, const char *fsrc, std::initializer_list<const char *> attribs)
	{
		static unsigned int id = 0;

		prog = glCreateProgram();
		vsh = glCreateShader(GL_VERTEX_SHADER);
		fsh = glCreateShader(GL_FRAGMENT_SHADER);

		if (!prog)
			Main::Exit("glCreateProgram() failed.");
		if (!(vsh && fsh))
			Main::Exit("glCreateShader() failed.");

		glAttachShader(prog, vsh);
		glAttachShader(prog, fsh);
		glShaderSource(vsh, 1, &vsrc, 0);
		glShaderSource(fsh, 1, &fsrc, 0);
		glCompileShader(vsh);
		glCompileShader(fsh);

		GLint vstat, fstat;
		glGetShaderiv(vsh, GL_COMPILE_STATUS, &vstat);
		glGetShaderiv(fsh, GL_COMPILE_STATUS, &fstat);

		if (vstat != GL_TRUE || fstat != GL_TRUE)
		{
			char *vlog, *flog;
			glGetShaderiv(vsh, GL_INFO_LOG_LENGTH, &vstat);
			glGetShaderiv(fsh, GL_INFO_LOG_LENGTH, &fstat);
			if (vstat)
			{
				vlog = new char[vstat];
				glGetShaderInfoLog(vsh, vstat, 0, vlog);
			}
			else
				vlog = (char *) "<empty>";
			if (fstat)
			{
				flog = new char[fstat];
				glGetShaderInfoLog(fsh, fstat, 0, flog);
			}
			else
				flog = (char *) "<empty>";

			Main::Exit(Jo("Shader #", id, " compilation failed:",
						  (vstat != GL_TRUE) ? "\nVertex shader: NOT OK":"\nVertex shader: OK",
						  (fstat != GL_TRUE) ? "\nFragment shader: NOT OK":"\nFragment shader: OK",
						  "\n\n  Vertex log:\n", vlog, "\n  Fragment log:\n", flog));
		}

		int loc = 0;
		for (const char *it : attribs)
		{
			if (it)
				glBindAttribLocation(prog, loc++, it);
		}
		glLinkProgram(prog);
		glGetProgramiv(prog, GL_LINK_STATUS, &vstat);
		if (vstat != GL_TRUE)
		{
			glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &vstat);
			char *log;
			if (vstat)
			{
				log = new char[vstat];
				glGetProgramInfoLog(prog, vstat, 0, log);
			}
			else
				log = (char *) "<empty>";

			Main::Exit(Jo("Shader #", id, " linking failed.\n\n  Log:\n",log));
		}

		id++;
	}

	const float (*const CameraController::r)[3] = M::Quaternion::GetMassTransformMatrix3x3();

	GLuint BindingContainer::vao_binding = 0;
	GLuint BindingContainer::vbo_binding = 0;
	#if !(!defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2)
	int BindingContainer::attr_c = 0;
	#endif

	ImageData::ImageData(const char *fname)
	{
		std::ifstream inp_file(fname, std::ios_base::in | std::ios_base::binary);
		if (!inp_file)
			Main::Exit(Jo("Can't load ", fname, ": File not found."));
		inp_file >> std::noskipws;
		unsigned char idfieldlen, tmp;
		bool swap_x, swap_y;

		inp_file >> idfieldlen;																							// 00
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Jo("Can't load ", fname, ": Files with colormaps are not supported.")); // 01
		inp_file >> tmp; if (tmp != 0x02) Main::Exit(Jo("Can't load ", fname, ": File must uncompressed 32 bpp TrueColor image (type check failure).")); // 02
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Jo("Can't load ", fname, ": Files with colormaps are not supported (1).")); // 03
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Jo("Can't load ", fname, ": Files with colormaps are not supported (2).")); // 04
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Jo("Can't load ", fname, ": Files with colormaps are not supported (3).")); // 05
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Jo("Can't load ", fname, ": Files with colormaps are not supported (4).")); // 06
		inp_file >> tmp; if (tmp != 0x00) Main::Exit(Jo("Can't load ", fname, ": Files with colormaps are not supported (5).")); // 07
		inp_file.ignore(4);                                                                                             // 08
																														// 09
																														// 0a
																														// 0b
		inp_file >> tmp; img_w = tmp;																					// 0c
		inp_file >> tmp; img_w |= tmp << 8;																				// 0d
		inp_file >> tmp; img_h = tmp;																					// 0e
		inp_file >> tmp; img_h |= tmp << 8;																				// 0f
		inp_file >> tmp; if (tmp != 0x20) Main::Exit(Jo("Can't load ", fname, ": File must uncompressed 32 bpp TrueColor image (bpp check failure).")); // 10
		inp_file >> tmp; if ((tmp&15)!=8) Main::Exit(Jo("Can't load ", fname, ": File must uncompressed 32 bpp TrueColor image (alpha check failure).")); // 11

		swap_x = tmp & 0x10; swap_y = !(tmp & 0x20);

		inp_file.ignore(idfieldlen);

		if (!inp_file) Main::Exit(Jo("Can't load ", fname, ": Unexpected end of the file (header)."));

		data = new unsigned char[img_w * img_h][4];
		int x, y, xend, yend, xstep, ystep, xbegin;

		if (!swap_x)
		{
			xbegin = 0;
			xend = img_w;
			xstep = 1;
		}
		else
		{
			xbegin = img_w - 1;
			xend = -1;
			xstep = -1;
		}
		if (!swap_y)
		{
			y = 0;
			yend = img_h;
			ystep = 1;
		}
		else
		{
			y = img_h - 1;
			yend = -1;
			ystep = -1;
		}


		while (y != yend)
		{
			x = xbegin;
			while (x != xend)
			{
				auto &ref = data[x + img_w * y];
				inp_file >> ref[2];
				inp_file >> ref[1];
				inp_file >> ref[0];
				inp_file >> ref[3];
				x += xstep;
			}
			y += ystep;
		}
		if (!inp_file) Main::Exit(Jo("Can't load ", fname, ": Unexpected end of the file."));

		bytelen = img_w * img_h * 4 * sizeof (unsigned char);
	}

	void Depth(bool on)
	{
		if (on)
		{
			glEnable(GL_DEPTH_TEST);
			bufclear_flags = GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT;
		}
		else
		{
			glDisable(GL_DEPTH_TEST);
			bufclear_flags = GL_COLOR_BUFFER_BIT;
		}
	}

	void Init()
	{
		EXEC_ONCE;

		static char openglfailstring[] = "OpenGL ?.? "
										 #if MANIFEST_OPENGL_PROFILE == 0
										 "(core profile)"
										 #elif MANIFEST_OPENGL_PROFILE == 1
										 "(compatibility profile)"
										 #else
										 "ES"
										 #endif
										 " context init failed.\n"
										 #if MANIFEST_OPENGL_PROFILE != 2
										 "\n\n\n"
										 "1. Update your videocard driver.\n"
										 "2. Force your driver to use best available videocard for this application.\n"
										 "If it still complains about your OpenGL, then your videocard\n"
										 "is probably too old for this program.\n"
										 #else
										 "Your device or firmware is too old."
										 #endif
										 ;

		static_assert(!(MANIFEST_OPENGL_MSAA & (MANIFEST_OPENGL_MSAA - 1)) && MANIFEST_OPENGL_MSAA >= 0 && MANIFEST_OPENGL_MSAA <= 16, "MSAA is not valid!");

		openglfailstring[7] = MANIFEST_OPENGL_MAJOR + '0';
		openglfailstring[9] = MANIFEST_OPENGL_MINOR + '0';

		#if !defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2
		glewExperimental = 1;
		#endif

		if (MANIFEST_OPENGL_MSAA > 1)
		{
			SDL_GL_SetAttribute(SDL_GL_MULTISAMPLEBUFFERS, 1);
			SDL_GL_SetAttribute(SDL_GL_MULTISAMPLESAMPLES, MANIFEST_OPENGL_MSAA);
		}

		if (!(conhandle = SDL_GL_CreateContext(Window::Handle())))
			Main::Exit(openglfailstring);

		SDL_GL_SetSwapInterval(MANIFEST_SYSTEM_VSYNC ? 1 : 0);

		#if !defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2
		if (glewInit() != GLEW_OK)
			Main::Exit("GLEW init failed.");
		#endif

		glEnable(GL_BLEND);
		glDepthFunc(GL_LEQUAL);
		//glDepthFunc(GL_GEQUAL);
		//glClearDepthf(-1);
	}

	void PreTick()
	{
		glClear(bufclear_flags);
	}

	void PostTick()
	{
		SDL_GL_SwapWindow(Window::Handle());
		ForceErrorCheck();
	}
}
