#ifndef GFX_H_INCL
#define GFX_H_INCL

#include <cstring>
#include <iostream>
#include <list>
#include <type_traits>
#include "libsdl.h"

#include "system.h"
#include "utils.h"

#include "libogl.h"

//   Perspective

//              +Z               //
//                               //
//            _______            //
//           /  +X   \           //
//  +Y      /         \     -Y   //
//         /           \         //
//        /     -X      \        //
//        """""""""""""""        //
//                               //
//              -Z               //


//   Flat

//                               //
//                               //
//              -Y               //
//         +-----------+         //
//         |'.       .'|         //
//         |  '+ - +'  |         //
//     -X  |   : 0 :   | +X      //
//         |  .+ - +.  |         //
//         |.'  +Z   '.|         //
//         +-----------+         //
//              +Y               //
//                               //
//              -Z               //

namespace GFX
{
	void ForceErrorCheck();

	SDL_GLContext ConHandle();

	void Depth(bool on);
	inline void Culling(bool on) {(on ? glEnable : glDisable)(GL_CULL_FACE);}
	inline void Blending(bool on) {(on ? glEnable : glDisable)(GL_BLEND);}
	inline void NormalBlendFunc() {glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA);}
	inline void OverwritingBlendFunc() {glBlendFunc(GL_ONE, GL_ZERO);}
	inline void ClearColor(M::vec4<unsigned char> c) {glClearColor(c.r / 255.0f, c.g / 255.0f, c.b / 255.0f, c.a / 255.0f);}
	inline void Viewport(M::vec2<int> pos, M::vec2<int> size) {glViewport(pos.x, pos.y, size.x, size.y);}
	inline void Viewport_Fullscreen() {glViewport(0, 0, Window::W(), Window::H());}
	#if !defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2
	inline void DepthClamp(bool on) {(on ? glEnable : glDisable)(GL_DEPTH_CLAMP);}
	#else
	[[deprecated("Has no effect on ES.")]]
	inline void DepthClamp(bool) {}
	#endif

	namespace BlendFunc
	{
		inline void Normal()
		{
			glBlendFuncSeparate(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA);
		}
		inline void None()
		{
			glBlendFunc(GL_ONE, GL_ZERO);
		}
	}

	inline namespace TmplDefs
	{
		template <typename T> struct type2glconst;
		template <> struct type2glconst<unsigned char>  {static constexpr GLenum value = GL_UNSIGNED_BYTE;};
		template <> struct type2glconst<signed char>    {static constexpr GLenum value = GL_BYTE;};
		template <> struct type2glconst<char>           {static constexpr GLenum value = Utils::is_char_signed ? GL_BYTE : GL_UNSIGNED_BYTE;};
		template <> struct type2glconst<unsigned short> {static constexpr GLenum value = GL_UNSIGNED_SHORT;};
		template <> struct type2glconst<signed short>   {static constexpr GLenum value = GL_SHORT;};
		template <> struct type2glconst<unsigned int>   {static constexpr GLenum value = GL_UNSIGNED_INT;};
		template <> struct type2glconst<signed int>     {static constexpr GLenum value = GL_INT;};
		template <> struct type2glconst<float>          {static constexpr GLenum value = GL_FLOAT;};
		#ifdef GL_DOUBLE
		template <> struct type2glconst<double>         {static constexpr GLenum value = GL_DOUBLE;};
		#endif

		#if !defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2
		template <typename T>
		inline void type2glattribfunc(GLuint index, GLint size, GLsizei stride, const GLvoid *pointer)
		{
			static_assert(type2glconst<T>::value, "Invalid type!");
			glVertexAttribIPointer(index, size, type2glconst<T>::value, stride, pointer);
		}
		template <>
		inline void type2glattribfunc<float>(GLuint index, GLint size, GLsizei stride, const GLvoid *pointer)
		{
			glVertexAttribPointer(index, size, type2glconst<float>::value, GL_FALSE, stride, pointer);
		}
		template <>
		inline void type2glattribfunc<double>(GLuint index, GLint size, GLsizei stride, const GLvoid *pointer)
		{
			glVertexAttribLPointer(index, size, type2glconst<double>::value, stride, pointer);
		}
		#else
		template <typename T>
		inline void type2glattribfunc(GLuint index, GLint size, GLsizei stride, const GLvoid *pointer)
		{
			static_assert(std::is_same<T, float>::value, "Invalid type!");
			glVertexAttribPointer(index, size, type2glconst<float>::value, GL_FALSE, stride, pointer);
		}
		#endif



		template <typename T, typename ...P>
		struct AttribLayout_InternalValue
		{
			static_assert(!std::is_same<T, void>::value, "Void vector is used in layout!");
			using curtype = T;
			curtype cur;
		};

		template <typename T, typename ...P>
		struct AttribLayout_InternalContainer
		{
			static_assert(!std::is_same<T, void>::value, "Void vector is used in layout!");
			using curtype = T;
			using nexttype = typename std::conditional<(sizeof...(P) > 1), AttribLayout_InternalContainer<P...>, AttribLayout_InternalValue<P...>>::type;
			curtype cur;
			nexttype next;
		};

		template <typename ...P>
		class AttribLayout
		{
		  protected:
			static_assert(sizeof...(P) > 0, "Zero-length attrib layout!");
			using cont_type = typename std::conditional<(sizeof...(P) > 1), AttribLayout_InternalContainer<P...>, AttribLayout_InternalValue<P...>>::type;
		  public:
			static constexpr int elements = sizeof...(P);
		  protected:
			cont_type storage;

			template <unsigned int N, typename T>
			struct type_at_internal
			{
				using type = typename type_at_internal<(N - 1), typename T::nexttype>::type;
			};
			template <typename T>
			struct type_at_internal<0, T>
			{
				using type = T;
			};

			template <unsigned int N> using cont_type_at = typename type_at_internal<N, cont_type>::type;

			template <unsigned int T, unsigned int N>
			struct at_internal
			{
				static cont_type_at<N - T> &func(cont_type &ref)
				{
					return at_internal<T + 1, N>::func(ref).next;
				}
			};

			template <unsigned int T>
			struct at_internal<T, T>
			{
				static cont_type_at<0> &func(cont_type &ref)
				{
					return ref;
				}
			};
		  public:
			template <unsigned int N> using type_at = typename type_at_internal<N, cont_type>::type::curtype;

			template <unsigned int N> type_at<N> &at() {return at_internal<0, N>::func(storage).cur;}

		  protected:
			template <unsigned int N, unsigned int M, typename F, typename ...L> struct ctor_internal
			{
				static void func(AttribLayout *ptr, const F &ref, const L &...pack)
				{
					ptr->at<N>() = ref;
					ctor_internal<N + 1, M, L...>::func(ptr, pack...);
				}
			};
			template <unsigned int M, typename F> struct ctor_internal<M, M, F>
			{
				static void func(AttribLayout *ptr, const F &ref)
				{
					ptr->at<M>() = ref;
				}
			};
		  public:
			AttribLayout() {}
			AttribLayout(const P &...params)
			{
				ctor_internal<0, elements - 1, P...>::func(this, params...);
			}
		};

		namespace InternalStuff
		{
			template <typename T, unsigned int N, unsigned int M, unsigned int O> struct SetVertexAttribs_internal
			{
				static void func()
				{
					glEnableVertexAttribArray(N);
					type2glattribfunc<typename T::template type_at<N>::type>(N, T::template type_at<N>::dim, sizeof (T), (void *)O);
					SetVertexAttribs_internal<T, N + 1, M, O + sizeof (typename T::template type_at<N>)>::func();
				}
			};
			template <typename T, unsigned int M, unsigned int O> struct SetVertexAttribs_internal<T, M, M, O>
			{
				static void func()
				{
					glEnableVertexAttribArray(M);
					type2glattribfunc<typename T::template type_at<0>::type>(M, T::template type_at<M>::dim, sizeof (T), (void *)O);
				}
			};
		}

		template <typename T>
		void SetVertexAttribs()
		{
			InternalStuff::SetVertexAttribs_internal<T, 0, T::elements - 1, 0>::func();
		}

		template <>
		inline void SetVertexAttribs<void>() {}

		template <typename T> struct AttribLayoutSize
		{
			static constexpr unsigned int value = T::elements;
		};
		template <> struct AttribLayoutSize<void>
		{
			static constexpr unsigned int value = 0;
		};
	}

	enum class Access : GLenum
	{
		draw_static  = GL_STATIC_DRAW,
		draw_dynamic = GL_DYNAMIC_DRAW,
		draw_stream  = GL_STREAM_DRAW,
		#if !defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2
		copy_static  = GL_STATIC_COPY,
		copy_dynamic = GL_DYNAMIC_COPY,
		copy_stream  = GL_STREAM_COPY,
		read_static  = GL_STATIC_READ,
		read_dynamic = GL_DYNAMIC_READ,
		read_stream  = GL_STREAM_READ,
		#else
		copy_static  = GL_STATIC_DRAW,
		copy_dynamic = GL_DYNAMIC_DRAW,
		copy_stream  = GL_STREAM_DRAW,
		read_static  = GL_STATIC_DRAW,
		read_dynamic = GL_DYNAMIC_DRAW,
		read_stream  = GL_STREAM_DRAW,
		#endif
	};

	template <typename> class VAO;
	class BindingContainer
	{
		~BindingContainer() = delete;
		static GLuint vbo_binding;
		static GLuint vao_binding;
		#if !(!defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2)
		static int attr_c;
		#endif
		template <typename> friend class VAO;
	};

	#if !defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2
	template <typename T> // T = AttribLayout or void
	class VAO
	{

		GLuint vao, vbo;
	  public:
		VAO(const T *data, unsigned int bytelen, Access ac_type = Access::draw_static) // (`data` may be null) Binds VAO & VBO after construction
		{
			glGenVertexArrays(1, &vao);
			glGenBuffers(1, &vbo);
			static unsigned id = 0;
			if (!vao)
				Main::Exit(Jo("glGenVertexArrays() failed for VAO #", id, '!'));
			if (!vbo)
				Main::Exit(Jo("glGenBuffers() failed for VAO #", id, '!'));
			Bind();
			VBO_Bind();

			glBufferData(GL_ARRAY_BUFFER, bytelen, data, (GLenum)ac_type);

			SetVertexAttribs<T>();

			id++;
		}

		VAO(const VAO &) = delete;
		VAO(VAO &&) = delete;
		VAO &operator=(const VAO &) = delete;
		VAO &operator=(VAO &&) = delete;

		~VAO()
		{
			glDeleteVertexArrays(1, &vao);
			glDeleteBuffers(1, &vbo);
		}

		void NewData(void *data, unsigned int bytelen, Access ac_type = Access::draw_static) // (`data` may be null) Auto binds VBO
		{
			VBO_Bind();
			glBufferData(GL_ARRAY_BUFFER, bytelen, data, (GLenum)ac_type);
		}
		void Bind() const
		{
			if (vao == BindingContainer::vao_binding)
				return;
			BindingContainer::vao_binding = vao;
			glBindVertexArray(vao);
		}
		static void BindNothing()
		{
			BindingContainer::vao_binding = 0;
			glBindVertexArray(0);
		}
		void VBO_Bind() const
		{
			if (vbo == BindingContainer::vbo_binding)
				return;
			BindingContainer::vbo_binding = vbo;
			glBindBuffer(GL_ARRAY_BUFFER, vbo);
		}
		static void VBO_BindNothing()
		{
			BindingContainer::vbo_binding = 0;
			glBindBuffer(GL_ARRAY_BUFFER, 0);
		}

		void Get(void *ptr, unsigned int pos, unsigned int size) const // Auto binds VBO
		{
			VBO_Bind();
			glGetBufferSubData(GL_ARRAY_BUFFER, pos, size, ptr);
		}

		void Set(const void *ptr, unsigned int pos, unsigned int size) // Auto binds VBO
		{
			VBO_Bind();
			glBufferSubData(GL_ARRAY_BUFFER, pos, size, ptr);
		}

		void Draw(unsigned int pos, unsigned int count) const // Auto binds VAO
		{
			Bind();
			glDrawArrays(GL_TRIANGLES, pos, count);
		}
		void DrawLines(unsigned int pos, unsigned int count) const // Auto binds VAO
		{
			Bind();
			glDrawArrays(GL_LINES, pos, count);
		}
		void DrawPoints(unsigned int pos, unsigned int count) const // Auto binds VAO
		{
			Bind();
			glDrawArrays(GL_POINTS, pos, count);
		}
		[[deprecated("Warning: VBO w/ quads is nonstandard GL extension. Use Draw() and Push4as3x2() instead.")]]
		void DrawQuads(unsigned int pos, unsigned int count) const // Auto binds VAO
		{
			Bind();
			glDrawArrays(GL_QUADS, pos, count);
		}
	};
	#else
	template <typename T> // DIFFERENCE = every call from class VAO that bind vao, here binds vbo instead
	class VAO
	{
		GLuint vao, vbo;
	  public:
		VAO(const T *data, unsigned int bytelen, Access ac_type = Access::draw_static) // (`data` may be null) Binds VAO & VBO after construction
		{
			glGenBuffers(1, &vbo);
			GLuint id = 1;
			if (!vbo)
				Main::Exit(Jo("glGenBuffers() failed for VAO #", id, '!'));
			Bind();
			VBO_Bind();

			glBufferData(GL_ARRAY_BUFFER, bytelen, data, (GLenum)ac_type);

			SetVertexAttribs<T>();
			vao = id;

			id++;
			if (!id)
				id = 1;
		}

		VAO(const VAO &) = delete;
		VAO(VAO &&) = delete;
		VAO &operator=(const VAO &) = delete;
		VAO &operator=(VAO &&) = delete;

		~VAO()
		{
			glDeleteBuffers(1, &vbo);
		}

		void NewData(void *data, unsigned int bytelen, Access ac_type = Access::draw_static) // (`data` may be null) Auto binds VBO
		{
			VBO_Bind();
			glBufferData(GL_ARRAY_BUFFER, bytelen, data, (GLenum)ac_type);
		}
		void Bind() const
		{
			if (vao == BindingContainer::vao_binding)
				return;
			BindingContainer::vao_binding = vao;
			for (int i = 0; i < BindingContainer::attr_c; i++)
				glDisableVertexAttribArray(i);
			BindingContainer::attr_c = AttribLayoutSize<T>::value;
			VBO_Bind();
			SetVertexAttribs<T>();
		}
		static void BindNothing()
		{
			BindingContainer::vao_binding = 0;
			VBO_BindNothing();
		}
		void VBO_Bind() const
		{
			if (vbo == BindingContainer::vbo_binding)
				return;
			BindingContainer::vbo_binding = vbo;
			glBindBuffer(GL_ARRAY_BUFFER, vbo);
		}
		static void VBO_BindNothing()
		{
			BindingContainer::vbo_binding = 0;
			glBindBuffer(GL_ARRAY_BUFFER, 0);
		}

		#if !defined(MANIFEST_OPENGL_PROFILE) || MANIFEST_OPENGL_PROFILE != 2
		void Get(void *ptr, unsigned int pos, unsigned int size) const // Auto binds VBO
		{
			VBO_Bind();
			glGetBufferSubData(GL_ARRAY_BUFFER, pos, size, ptr);
		}
		#else
		[[deprecated("WARNING: DOES NOT WORK ON ES")]]
		void Get(void *, unsigned int, unsigned int) const {}
		#endif

		void Set(const void *ptr, unsigned int pos, unsigned int size) // Auto binds VBO
		{
			VBO_Bind();
			glBufferSubData(GL_ARRAY_BUFFER, pos, size, ptr);
		}

		void Draw(unsigned int pos, unsigned int count) const // Auto binds VAO
		{
			Bind();
			glDrawArrays(GL_TRIANGLES, pos, count);
		}
		void DrawLines(unsigned int pos, unsigned int count) const // Auto binds VAO
		{
			Bind();
			glDrawArrays(GL_LINES, pos, count);
		}
		void DrawPoints(unsigned int pos, unsigned int count) const // Auto binds VAO
		{
			Bind();
			glDrawArrays(GL_POINTS, pos, count);
		}
	};
	#endif

	class ImageData
	{
		unsigned int img_w, img_h, bytelen;
		unsigned char (*data)[4];
	  public:
	  	unsigned int  W()       const {return img_w;}
	  	unsigned int  H()       const {return img_h;}
	  	unsigned char (*Data() const)[4] {return data;}
	  	unsigned int  ByteLen() const {return bytelen;}

		ImageData(const char *fname);

		ImageData(unsigned int w, unsigned int h)
		{
			data = new unsigned char[w * h][4];
			img_w = w;
			img_h = h;
			bytelen = w * h * 4 * sizeof (unsigned char);
		}
		ImageData(const ImageData &o)
		{
			img_w = o.img_w;
			img_h = o.img_h;
			bytelen = o.bytelen;
			data = new unsigned char[img_w * img_h][4];
			std::memcpy(data, o.data, bytelen);
		}
		ImageData(ImageData &&o)
		{
			img_w = o.img_w;
			img_h = o.img_h;
			bytelen = o.bytelen;
			data = o.data;
			o.data = 0;
		}
		ImageData &operator=(const ImageData &o)
		{
			delete [] data;
			img_w = o.img_w;
			img_h = o.img_h;
			bytelen = o.bytelen;
			data = new unsigned char[img_w * img_h][4];
			std::memcpy(data, o.data, bytelen);
			return *this;
		}
		ImageData &operator=(ImageData &&o)
		{
			delete [] data;
			img_w = o.img_w;
			img_h = o.img_h;
			bytelen = o.bytelen;
			data = o.data;
			o.data = 0;
			return *this;
		}

		~ImageData()
		{
			if (data)
				delete [] data;
		}
	};

	class Texture
	{
		GLuint tex;
	  public:
	  	enum class Interpolation : GLenum
		{
			nearest = GL_NEAREST,
			linear = GL_LINEAR,
		};

		enum class Wrap : GLenum
		{
			clamp = GL_CLAMP_TO_EDGE,
			mirror = GL_MIRRORED_REPEAT,
			repeat = GL_REPEAT,
			#if MANIFEST_OPENGL_PROFILE != 2 && (MANIFEST_OPENGL_MAJOR > 1 || MANIFEST_OPENGL_MINOR >= 2)
			transp = GL_CLAMP_TO_BORDER,
			#else
			transp = GL_CLAMP_TO_EDGE
			#endif
		};

		Texture()
		{
			glGenTextures(1, &tex);
			if (!tex)
				Main::Exit("glGenTextures() failed.");
		}
		Texture(const Texture &) = delete;
		Texture(Texture &&) = delete;
		Texture &operator=(const Texture &) = delete;
		Texture &operator=(Texture &&) = delete;
		~Texture()
		{
			glDeleteTextures(1, &tex);
		};

		static void SetActiveSlot(unsigned int n)
		{
			glActiveTexture(GL_TEXTURE0 + n);
		}
		void BindToActiveSlot()
		{
			glBindTexture(GL_TEXTURE_2D, tex);
		}
		static void BindNothingToActiveSlot()
		{
			glBindTexture(GL_TEXTURE_2D, 0);
		}
		static void SetSlotAsUniform(GLint loc, unsigned int n)
		{
			glUniform1i(loc, n);
		}
		static void ActiveSlotParams(Interpolation itp, Wrap wrp)
		{
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, (GLuint)itp);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, (GLuint)itp);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, (GLuint)wrp);
			glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, (GLuint)wrp);
		}

		static void ActiveSlotData(const void *data, unsigned int w, unsigned int h) // `data` may be null
		{
			#ifdef GL_RGBA8
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
			#else
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_RGBA, GL_UNSIGNED_BYTE, data);
			#endif
		}
		static void ActiveSlotData(const ImageData &ref)
		{
			#ifdef GL_RGBA8
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, ref.W(), ref.H(), 0, GL_RGBA, GL_UNSIGNED_BYTE, ref.Data());
			#else
			glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, ref.W(), ref.H(), 0, GL_RGBA, GL_UNSIGNED_BYTE, ref.Data());
			#endif
		}
	};

	template <typename L> // L = AttribLayout
	class RenderingQueue
	{
	  protected:
		static_assert(!std::is_void<L>::value, "Void is not allowed as rendering queue type.");

		VAO<L> vao;
		unsigned int size;
		unsigned int pos;
		L *arr;

	  public:
		RenderingQueue(unsigned int l, Access acc = Access::draw_dynamic) : vao(0, sizeof (L) * l, acc)
		{
			if (l == 0)
				Main::Exit("Invalid rendering queue size.");
			size = l;
			pos = 0;
			arr = new L[l];
		}
		RenderingQueue(const RenderingQueue &) = delete;
		RenderingQueue(RenderingQueue &&) = delete;
		RenderingQueue &operator=(const RenderingQueue &) = delete;
		RenderingQueue &operator=(RenderingQueue &&) = delete;

		~RenderingQueue()
		{
			delete [] arr;
		}

		void ChangeSize(unsigned int l)
		{
			if (l == 0)
				Main::Exit("Invalid rendering queue size.");
			size = l;
			pos = 0;
			delete [] arr;
			arr = new L[l];
			vao.~VAO();
			new(&vao) VAO<L>(0, sizeof (L) * l, Access::draw_dynamic);
		}

		void Push(const L &x)
		{
			if (pos >= size)
				Main::Exit("Rendering queue overflow!");
			arr[pos    ] = x;
			pos += 1;
		}
		void Push2(const L &x, const L &y)
		{
			if (pos + 1 >= size)
				Main::Exit("Rendering queue overflow!");
			arr[pos    ] = x;
			arr[pos + 1] = y;
			pos += 2;
		}
		void Push3(const L &x, const L &y, const L &z)
		{
			if (pos + 2 >= size)
				Main::Exit("Rendering queue overflow!");
			arr[pos    ] = x;
			arr[pos + 1] = y;
			arr[pos + 2] = z;
			pos += 3;
		}
		[[deprecated("Warning: VBO w/ quads is nonstandard GL extension. Use Draw() and Push4as3x2() instead.")]]
		void Push4(const L &x, const L &y, const L &z, const L &w)
		{
			if (pos + 3 >= size)
				Main::Exit("Rendering queue overflow!");
			arr[pos    ] = x;
			arr[pos + 1] = y;
			arr[pos + 2] = z;
			arr[pos + 3] = w;
			pos += 4;
		}
		void Push4as3x2(const L &x, const L &y, const L &z, const L &w) // a b d  b c d
		{
			if (pos + 5 >= size)
				Main::Exit("Rendering queue overflow!");
			arr[pos    ] = x;
			arr[pos + 1] = y;
			arr[pos + 2] = w;
			arr[pos + 3] = y;
			arr[pos + 4] = z;
			arr[pos + 5] = w;
			pos += 6;
		}

		void Send() // You shall not call Push*() between Send() and Draw*()
		{
			vao.Set(arr, 0, pos * sizeof (L));
		}
		void Clear()
		{
			pos = 0;
		}

		void Draw()
		{
			vao.Draw(0, pos);
		}
		void DrawPoints()
		{
			vao.DrawPoints(0, pos);
		}
		void DrawLines()
		{
			vao.DrawLines(0, pos);
		}
		#ifdef GL_QUADS
		void DrawQuads()
		{
			vao.DrawQuads(0, pos);
		}
		#endif
	};

	class Shader
	{
		GLuint prog, vsh, fsh;

	public:

		Shader(const char *vsrc, const char *fsrc, std::initializer_list<const char *> attribs);

		Shader(const Shader &) = delete;
		Shader(Shader &&) = delete;
		Shader &operator=(const Shader &) = delete;
		Shader &operator=(Shader &&) = delete;

		operator GLuint() const
		{
			return prog;
		}

		void Use() const
		{
			glUseProgram(prog);
		}
		static void UseNothing()
		{
			glUseProgram(0);
		}
		~Shader()
		{
			glDeleteShader(vsh);
			glDeleteShader(fsh);
			glDeleteProgram(prog);
		}
	};

	class CameraController
	{
		#define OPT_MAT 1 // Disable some of useless matrix computations
		// r00 r10 r20  0    - model matrix
		// r01 r11 r21  0
		// r02 t12 t22  0
		//  x   y   z   1

		  //  1   0   0   0    - campos matrix
		  //  0   1   0   0
		  //  0   0   1   0
		  // -x  -y  -z   1

		  // r00 r10 r20  0    - camrot matrix
		  // r01 r11 r21  0
		  // r02 t12 t22  0
		  //  0   0   0   1

		//              r00                              r10                              r20                 0    - camera matrix (campos * camrot)
		//              r01                              r11                              r21                 0
		//              r02                              r12                              r22                 0
		// -x * r00 + -y * r01 + -z * r02   -x * r10 + -y * r11 + -z * r12   -x * r20 + -y * r21 + -z * r22   1

		//  0   0  mzz  persp_dist_comp    - view matrix (perspective)
		// mxx  0   0          0
		//  0  myy  0          0
		//  0   0  mzw         0

		// pe_[0]1 = mxx = -myy / w_div_h
		// pe12 = myy = std::tan((fPi - y_view_angle) / 2.0f);
		// pe20 = mzz = (znear + zfar) / (znear - zfar);
		// pe23 = 2.0f * zfar * znear / (zfar - znear);

		static const float (*const r)[3];
		float mo[4][4]; // Model                 matrix
		float ca[4][4]; //         Camera        matrix
		float pe[4][4]; //                  View matrix
		float mc[4][4]; // Model & Camera        matrix
		float cp[4][4]; //         Camera & View matrix
		float m [4][4]; // Model & Camera & View matrix - final
		float pe_[2][4], cp_[2][4], m_[2][4];

		static constexpr float persp_dist_comp = 1;

	  public:
		CameraController()
		{
			mo[3][0] = 0; mo[3][1] = 0; mo[3][2] = 0; mo[3][3] = 1;
			ca[3][0] = 0; ca[3][1] = 0; ca[3][2] = 0; ca[3][3] = 1;
			mc[3][0] = 0; mc[3][1] = 0; mc[3][2] = 0; mc[3][3] = 1;
		}

		void SetUniModel(GLint loc) const
		{
			glUniformMatrix4fv(loc, 1, GL_FALSE, (float *)mo);
		}
		void SetUniCam(GLint loc) const
		{
			glUniformMatrix4fv(loc, 1, GL_FALSE, (float *)ca);
		}
		void SetUniView(GLint loc) const
		{
			glUniformMatrix4fv(loc, 1, GL_FALSE, (float *)pe);
		}
		void SetUniModelCam(GLint loc) const
		{
			glUniformMatrix4fv(loc, 1, GL_FALSE, (float *)mc);
		}
		void SetUniCamView(GLint loc) const
		{
			glUniformMatrix4fv(loc, 1, GL_FALSE, (float *)cp);
		}
		void SetUniFinal(GLint loc) const
		{
			glUniformMatrix4fv(loc, 1, GL_FALSE, (float *)m);
		}

		void ComputeModelCam()
		{
			mc[0][0] = Utils::ScalarFromMat4Product<0, 0>(mo, ca);
			mc[1][0] = Utils::ScalarFromMat4Product<1, 0>(mo, ca);
			mc[2][0] = Utils::ScalarFromMat4Product<2, 0>(mo, ca);
			#if OPT_MAT == 0
			mc[3][0] = Utils::ScalarFromMat4Product<3, 0>(mo, ca);
			#endif
			mc[0][1] = Utils::ScalarFromMat4Product<0, 1>(mo, ca);
			mc[1][1] = Utils::ScalarFromMat4Product<1, 1>(mo, ca);
			mc[2][1] = Utils::ScalarFromMat4Product<2, 1>(mo, ca);
			#if OPT_MAT == 0
			mc[3][1] = Utils::ScalarFromMat4Product<3, 1>(mo, ca);
			#endif
			mc[0][2] = Utils::ScalarFromMat4Product<0, 2>(mo, ca);
			mc[1][2] = Utils::ScalarFromMat4Product<1, 2>(mo, ca);
			mc[2][2] = Utils::ScalarFromMat4Product<2, 2>(mo, ca);
			#if OPT_MAT == 0
			mc[3][2] = Utils::ScalarFromMat4Product<3, 2>(mo, ca);
			#endif
			mc[0][3] = Utils::ScalarFromMat4Product<0, 3>(mo, ca);
			mc[1][3] = Utils::ScalarFromMat4Product<1, 3>(mo, ca);
			mc[2][3] = Utils::ScalarFromMat4Product<2, 3>(mo, ca);
			#if OPT_MAT == 0
			mc[3][3] = Utils::ScalarFromMat4Product<3, 3>(mo, ca);
			#endif
		}
		void ComputeCamView()
		{
			cp_[0][0] = Utils::ScalarFromMat4Product<0, 0>(ca, pe_);
			cp_[1][0] = Utils::ScalarFromMat4Product<1, 0>(ca, pe_);
			cp[2][0] = Utils::ScalarFromMat4Product<2, 0>(ca, pe);
			cp[3][0] = Utils::ScalarFromMat4Product<3, 0>(ca, pe);
			cp_[0][1] = Utils::ScalarFromMat4Product<0, 1>(ca, pe_);
			cp_[1][1] = Utils::ScalarFromMat4Product<1, 1>(ca, pe_);
			cp[2][1] = Utils::ScalarFromMat4Product<2, 1>(ca, pe);
			cp[3][1] = Utils::ScalarFromMat4Product<3, 1>(ca, pe);
			cp_[0][2] = Utils::ScalarFromMat4Product<0, 2>(ca, pe_);
			cp_[1][2] = Utils::ScalarFromMat4Product<1, 2>(ca, pe_);
			cp[2][2] = Utils::ScalarFromMat4Product<2, 2>(ca, pe);
			cp[3][2] = Utils::ScalarFromMat4Product<3, 2>(ca, pe);
			cp_[0][3] = Utils::ScalarFromMat4Product<0, 3>(ca, pe_);
			cp_[1][3] = Utils::ScalarFromMat4Product<1, 3>(ca, pe_);
			cp[2][3] = Utils::ScalarFromMat4Product<2, 3>(ca, pe);
			cp[3][3] = Utils::ScalarFromMat4Product<3, 3>(ca, pe);
		}
		void ComputeFinalUsingModelCamPrecalc()
		{
			m_[0][0] = Utils::ScalarFromMat4Product<0, 0>(mc, pe_);
			m_[1][0] = Utils::ScalarFromMat4Product<1, 0>(mc, pe_);
			m[2][0] = Utils::ScalarFromMat4Product<2, 0>(mc, pe);
			m[3][0] = Utils::ScalarFromMat4Product<3, 0>(mc, pe);
			m_[0][1] = Utils::ScalarFromMat4Product<0, 1>(mc, pe_);
			m_[1][1] = Utils::ScalarFromMat4Product<1, 1>(mc, pe_);
			m[2][1] = Utils::ScalarFromMat4Product<2, 1>(mc, pe);
			m[3][1] = Utils::ScalarFromMat4Product<3, 1>(mc, pe);
			m_[0][2] = Utils::ScalarFromMat4Product<0, 2>(mc, pe_);
			m_[1][2] = Utils::ScalarFromMat4Product<1, 2>(mc, pe_);
			m[2][2] = Utils::ScalarFromMat4Product<2, 2>(mc, pe);
			m[3][2] = Utils::ScalarFromMat4Product<3, 2>(mc, pe);
			m_[0][3] = Utils::ScalarFromMat4Product<0, 3>(mc, pe_);
			m_[1][3] = Utils::ScalarFromMat4Product<1, 3>(mc, pe_);
			m[2][3] = Utils::ScalarFromMat4Product<2, 3>(mc, pe);
			m[3][3] = Utils::ScalarFromMat4Product<3, 3>(mc, pe);
		}
		void ComputeFinalUsingCamViewPrecalc()
		{
			m_[0][0] = Utils::ScalarFromMat4Product<0, 0>(mo, cp_);
			m_[1][0] = Utils::ScalarFromMat4Product<1, 0>(mo, cp_);
			m[2][0] = Utils::ScalarFromMat4Product<2, 0>(mo, cp);
			m[3][0] = Utils::ScalarFromMat4Product<3, 0>(mo, cp);
			m_[0][1] = Utils::ScalarFromMat4Product<0, 1>(mo, cp_);
			m_[1][1] = Utils::ScalarFromMat4Product<1, 1>(mo, cp_);
			m[2][1] = Utils::ScalarFromMat4Product<2, 1>(mo, cp);
			m[3][1] = Utils::ScalarFromMat4Product<3, 1>(mo, cp);
			m_[0][2] = Utils::ScalarFromMat4Product<0, 2>(mo, cp_);
			m_[1][2] = Utils::ScalarFromMat4Product<1, 2>(mo, cp_);
			m[2][2] = Utils::ScalarFromMat4Product<2, 2>(mo, cp);
			m[3][2] = Utils::ScalarFromMat4Product<3, 2>(mo, cp);
			m_[0][3] = Utils::ScalarFromMat4Product<0, 3>(mo, cp_);
			m_[1][3] = Utils::ScalarFromMat4Product<1, 3>(mo, cp_);
			m[2][3] = Utils::ScalarFromMat4Product<2, 3>(mo, cp);
			m[3][3] = Utils::ScalarFromMat4Product<3, 3>(mo, cp);
		}
		void ComputeFinalWithoutPrecalc()
		{
			ComputeModelCam();
			ComputeFinalUsingModelCamPrecalc();
		}

		void ApplyRatioToMatrixes(float w) // Affects only view and derived matrixes
		{
			pe[0][0] = pe_[0][0] / w;
			pe[0][1] = pe_[0][1] / w;
			pe[0][2] = pe_[0][2] / w;
			pe[0][3] = pe_[0][3] / w;
			cp[0][0] = cp_[0][0] / w;
			cp[0][1] = cp_[0][1] / w;
			cp[0][2] = cp_[0][2] / w;
			cp[0][3] = cp_[0][3] / w;
			m[0][0]  = m_[0] [0] / w;
			m[0][1]  = m_[0] [1] / w;
			m[0][2]  = m_[0] [2] / w;
			m[0][3]  = m_[0] [3] / w;

			pe[1][0] = pe_[1][0];
			pe[1][1] = pe_[1][1];
			pe[1][2] = pe_[1][2];
			pe[1][3] = pe_[1][3];
			cp[1][0] = cp_[1][0];
			cp[1][1] = cp_[1][1];
			cp[1][2] = cp_[1][2];
			cp[1][3] = cp_[1][3];
			m[1][0]  = m_[1] [0];
			m[1][1]  = m_[1] [1];
			m[1][2]  = m_[1] [2];
			m[1][3]  = m_[1] [3];
		}
		void ApplyRatioToMatrixes(float w, float h) // Affects only view and derived matrixes
		{
			pe[0][0] = pe_[0][0] / w;
			pe[0][1] = pe_[0][1] / w;
			pe[0][2] = pe_[0][2] / w;
			pe[0][3] = pe_[0][3] / w;
			cp[0][0] = cp_[0][0] / w;
			cp[0][1] = cp_[0][1] / w;
			cp[0][2] = cp_[0][2] / w;
			cp[0][3] = cp_[0][3] / w;
			m[0][0]  = m_[0] [0] / w;
			m[0][1]  = m_[0] [1] / w;
			m[0][2]  = m_[0] [2] / w;
			m[0][3]  = m_[0] [3] / w;

			pe[1][0] = pe_[1][0] / h;
			pe[1][1] = pe_[1][1] / h;
			pe[1][2] = pe_[1][2] / h;
			pe[1][3] = pe_[1][3] / h;
			cp[1][0] = cp_[1][0] / h;
			cp[1][1] = cp_[1][1] / h;
			cp[1][2] = cp_[1][2] / h;
			cp[1][3] = cp_[1][3] / h;
			m[1][0]  = m_[1] [0] / h;
			m[1][1]  = m_[1] [1] / h;
			m[1][2]  = m_[1] [2] / h;
			m[1][3]  = m_[1] [3] / h;
		}

		void ModelPositionRotation(M::Vec<float, 3> p)
		{
			mo[0][0] = 1;
			mo[1][0] = 0;
			mo[2][0] = 0;
			#if OPT_MAT == 0
			mo[3][0] = 0;
			#endif

			mo[0][1] = 0;
			mo[1][1] = 1;
			mo[2][1] = 0;
			#if OPT_MAT == 0
			mo[3][1] = 0;
			#endif

			mo[0][2] = 0;
			mo[1][2] = 0;
			mo[2][2] = 1;
			#if OPT_MAT == 0
			mo[3][2] = 0;
			#endif

			mo[0][3] = p.x;
			mo[1][3] = p.y;
			mo[2][3] = p.z;
			#if OPT_MAT == 0
			mo[3][3] = 1;
			#endif
		}
		void ModelPositionRotation(M::Vec<float, 3> p, float scale)
		{
			mo[0][0] = scale;
			mo[1][0] = 0;
			mo[2][0] = 0;
			#if OPT_MAT == 0
			mo[3][0] = 0;
			#endif

			mo[0][1] = 0;
			mo[1][1] = scale;
			mo[2][1] = 0;
			#if OPT_MAT == 0
			mo[3][1] = 0;
			#endif

			mo[0][2] = 0;
			mo[1][2] = 0;
			mo[2][2] = scale;
			#if OPT_MAT == 0
			mo[3][2] = 0;
			#endif

			mo[0][3] = p.x;
			mo[1][3] = p.y;
			mo[2][3] = p.z;
			#if OPT_MAT == 0
			mo[3][3] = 1;
			#endif
		}
		void ModelPositionRotation(M::Vec<float, 3> p, const M::Quaternion &q)
		{
			q.PrepareMassTransform();

			mo[0][0] = r[0][0];
			mo[1][0] = r[1][0];
			mo[2][0] = r[2][0];
			#if OPT_MAT == 0
			mo[3][0] = 0;
			#endif

			mo[0][1] = r[0][1];
			mo[1][1] = r[1][1];
			mo[2][1] = r[2][1];
			#if OPT_MAT == 0
			mo[3][1] = 0;
			#endif

			mo[0][2] = r[0][2];
			mo[1][2] = r[1][2];
			mo[2][2] = r[2][2];
			#if OPT_MAT == 0
			mo[3][2] = 0;
			#endif

			mo[0][3] = p.x;
			mo[1][3] = p.y;
			mo[2][3] = p.z;
			#if OPT_MAT == 0
			mo[3][3] = 1;
			#endif
		}
		void ModelPositionRotation(M::Vec<float, 3> p, const M::Quaternion &q, float scale)
		{
			q.PrepareMassTransform();

			mo[0][0] = r[0][0] * scale;
			mo[1][0] = r[1][0] * scale;
			mo[2][0] = r[2][0] * scale;
			#if OPT_MAT == 0
			mo[3][0] = 0;
			#endif

			mo[0][1] = r[0][1] * scale;
			mo[1][1] = r[1][1] * scale;
			mo[2][1] = r[2][1] * scale;
			#if OPT_MAT == 0
			mo[3][1] = 0;
			#endif

			mo[0][2] = r[0][2] * scale;
			mo[1][2] = r[1][2] * scale;
			mo[2][2] = r[2][2] * scale;
			#if OPT_MAT == 0
			mo[3][2] = 0;
			#endif

			mo[0][3] = p.x;
			mo[1][3] = p.y;
			mo[2][3] = p.z;
			#if OPT_MAT == 0
			mo[3][3] = 1;
			#endif
		}
		void CameraPositionRotation(M::Vec<float, 3> p, const M::Quaternion &q)
		{
			p = -p;

			q.PrepareMassTransform();

			ca[0][0] = r[0][0];
			ca[1][0] = r[1][0];
			ca[2][0] = r[2][0];
			#if OPT_MAT == 0
			ca[3][0] = 0;
			#endif

			ca[0][1] = r[0][1];
			ca[1][1] = r[1][1];
			ca[2][1] = r[2][1];
			#if OPT_MAT == 0
			ca[3][1] = 0;
			#endif

			ca[0][2] = r[0][2];
			ca[1][2] = r[1][2];
			ca[2][2] = r[2][2];
			#if OPT_MAT == 0
			ca[3][2] = 0;
			#endif

			ca[0][3] = p.x * r[0][0] + p.y * r[0][1] + p.z * r[0][2];
			ca[1][3] = p.x * r[1][0] + p.y * r[1][1] + p.z * r[1][2];
			ca[2][3] = p.x * r[2][0] + p.y * r[2][1] + p.z * r[2][2];
			#if OPT_MAT == 0
			ca[3][3] = 1;
			#endif
		}
		void ViewParams_Perspective(float znear, float zfar, float y_view_angle) // angle is in radians
		{
			pe_[1][2] = std::tan((fPi - y_view_angle) / 2.0f);
			pe[2][0] = (znear + zfar) / (zfar - znear);
			pe[2][3] = 2.0f * zfar * znear / (znear - zfar);
			pe_[0][1] = -pe_[1][2];
			pe_[0][0] = pe_[0][2] = pe_[0][3] = 0;
			pe_[1][0] = pe_[1][1] = pe_[1][3] = 0;
			pe[2][1] = pe[2][2] = 0;
			pe[3][0] = persp_dist_comp;
			pe[3][1] = pe[3][2] = pe[3][3] = 0;
		}
		void ViewParams_Flat(float znear, float zfar) // WARNING: SUCKS FOR PIXEL GRAPHICS ON NON-MULTIPLY-OF-2 RESOLUTIONS BECAUSE THEN CENTER IS PLACED WITH HALF-PIXEL OFFSET
		{
			pe_[0][0] = 2; pe_[1][0] = 0; pe[2][0] = 0; pe[3][0] = 0;
			pe_[0][1] = 0; pe_[1][1] =-2; pe[2][1] = 0; pe[3][1] = 0;
			pe_[0][2] = 0; pe_[1][2] = 0; pe[2][2] = 2.0f / (zfar - znear); pe[3][2] = 0;
			pe_[0][3] = 0; pe_[1][3] = 0; pe[2][3] = (znear + zfar) / (znear - zfar); pe[3][3] = 1;
		}
		#undef OPT_MAT
	};
}

#endif

#if !defined(GFX_H_INCL_MAINT) && defined(GFX_H_MAINT_ACCESS)
#define GFX_H_INCL_MAINT
#undef GFX_H_MAINT_ACCESS
namespace GFX
{
	void Init();
	void PreTick();
	void PostTick();
}
#endif