#ifndef UTILS_H_INCL
#define UTILS_H_INCL

#include <cmath>
#include <cstdio>
#include <cstring>
#include <chrono>
#include <iostream>
#include <random>
#include <sstream>
#include <type_traits>
#include <utility>
#include "_manifest.h"
#include "main.h"

#define EXEC_ONCE \
do {static bool EXEC_ONCE_FLAG = 0; if (EXEC_ONCE_FLAG) Main::Exit(Jo("Error: ", __func__, "() called twice.")); EXEC_ONCE_FLAG = 1;} while (0)

#define TAG_TYPE(name) \
class name {~name() = delete;}

extern const float fPi;
extern const double dPi;
extern const float fPi2;
extern const double dPi2;
extern const float fPi_half;
extern const double dPi_half;
extern const float fPi_quarter;
extern const double dPi_quarter;

inline float deg2rad(float x) {return x * fPi / 180.0f;}
inline float rad2deg(float x) {return x * 180.0f / fPi;}
inline double deg2rad(double x) {return x * dPi / 180.0;}
inline double rad2deg(double x) {return x * 180.0 / dPi;}

namespace Tags
{
	// class Box
	TAG_TYPE(NoAutoCtor);     // Do not call obj ctor at container ctor, also adds .Ctor(...)
	TAG_TYPE(NoAutoDtor);     // Do not call obj dtor at container dtor, also adds .Dtor()
	TAG_TYPE(NoAutoCtorDtor); // 2 above together
	TAG_TYPE(ManLifetime);    // Adds .Ctor(...) and .Dtor()
	TAG_TYPE(Heap);           // Heap is used instead of in-class storage, heap space lifetime = container lifetime
	TAG_TYPE(Heap_DelUnused); // Heap is used instead of in-class storage, heap space lifetime = object lifetime
}

namespace Utils
{
	inline constexpr unsigned int DivCeil(unsigned int a, unsigned int b)
	{
		return (a + (b - 1)) / b ;
	}

	// * FP_Type
	// using FP_Type = ...; (double for everything except long double and long long, for them it's lond double)
	// * Vec
	// class Vec<Type, size>;  size = 1,2,3,4
	//   or
	// class vec1<type>;
	// class vec2<type>;
	// class vec3<type>;
	// class vec4<type>;
	//
	// * Quaternion
	// ...
	inline namespace Math
	{

		namespace Internal
		{
			template <typename T> struct Type
			{
				using type = double;
			};
			template <> struct Type<long double>
			{
				using type = long double;
			};
			template <> struct Type<long long>
			{
				using type = long double;
			};
		}
		template <typename T> using FP_Type = typename Internal::Type<T>::type;

		// START OF VEC DEFINITION

		#define COMMON static_assert(!std::is_lvalue_reference<T>::value && !std::is_rvalue_reference<T>::value, "Invalid type for a vector."); \
					   using type = T;                                                                                                          \
					   static constexpr bool is_integral = std::is_integral<T>::value;                                                          \
					   static constexpr bool is_floating = std::is_floating_point<T>::value;

		#define VAL(x) (*(x *)0)

		#define OP1_1(op)         template <typename T1> Vec<decltype(op VAL(T1)), 1> operator op(const Vec<T1, 1> &a     ) {return Vec<decltype(op a.x), 1>{(op a.x)};}
		#define OP2_1(op)         template <typename T1> Vec<decltype(op VAL(T1)), 2> operator op(const Vec<T1, 2> &a     ) {return Vec<decltype(op a.x), 2>{(op a.x), (op a.y)};}
		#define OP3_1(op)         template <typename T1> Vec<decltype(op VAL(T1)), 3> operator op(const Vec<T1, 3> &a     ) {return Vec<decltype(op a.x), 3>{(op a.x), (op a.y), (op a.z)};}
		#define OP4_1(op)         template <typename T1> Vec<decltype(op VAL(T1)), 4> operator op(const Vec<T1, 4> &a     ) {return Vec<decltype(op a.x), 4>{(op a.x), (op a.y), (op a.z), (op a.w)};}
		#define OP1_1_POSTFIX(op) template <typename T1> Vec<decltype(op VAL(T1)), 1> operator op(const Vec<T1, 1> &a, int) {return Vec<decltype(op a.x), 1>{(op a.x)};}
		#define OP2_1_POSTFIX(op) template <typename T1> Vec<decltype(op VAL(T1)), 2> operator op(const Vec<T1, 2> &a, int) {return Vec<decltype(op a.x), 2>{(op a.x), (op a.y)};}
		#define OP3_1_POSTFIX(op) template <typename T1> Vec<decltype(op VAL(T1)), 3> operator op(const Vec<T1, 3> &a, int) {return Vec<decltype(op a.x), 3>{(op a.x), (op a.y), (op a.z)};}
		#define OP4_1_POSTFIX(op) template <typename T1> Vec<decltype(op VAL(T1)), 4> operator op(const Vec<T1, 4> &a, int) {return Vec<decltype(op a.x), 4>{(op a.x), (op a.y), (op a.z), (op a.w)};}
		#define OP1_2(op)         template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 1> operator op(const Vec<T1, 1> &a, const Vec<T2, 1> &b) {return Vec<decltype(a.x op b.y), 1>{(a.x op b.x)};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 1> operator op(const Vec<T1, 1> &a, const     T2     &b) {return Vec<decltype(a.x op b  ), 1>{(a.x op b  )};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 1> operator op(const     T1     &a, const Vec<T2, 1> &b) {return Vec<decltype(a   op b.y), 1>{(a   op b.x)};}
		#define OP2_2(op)         template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 2> operator op(const Vec<T1, 2> &a, const Vec<T2, 2> &b) {return Vec<decltype(a.x op b.y), 2>{(a.x op b.x), (a.y op b.y)};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 2> operator op(const Vec<T1, 2> &a, const     T2     &b) {return Vec<decltype(a.x op b  ), 2>{(a.x op b  ), (a.y op b  )};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 2> operator op(const     T1     &a, const Vec<T2, 2> &b) {return Vec<decltype(a   op b.y), 2>{(a   op b.x), (a   op b.y)};}
		#define OP3_2(op)         template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 3> operator op(const Vec<T1, 3> &a, const Vec<T2, 3> &b) {return Vec<decltype(a.x op b.y), 3>{(a.x op b.x), (a.y op b.y), (a.z op b.z)};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 3> operator op(const Vec<T1, 3> &a, const     T2     &b) {return Vec<decltype(a.x op b  ), 3>{(a.x op b  ), (a.y op b  ), (a.z op b  )};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 3> operator op(const     T1     &a, const Vec<T2, 3> &b) {return Vec<decltype(a   op b.y), 3>{(a   op b.x), (a   op b.y), (a   op b.z)};}
		#define OP4_2(op)         template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 4> operator op(const Vec<T1, 4> &a, const Vec<T2, 4> &b) {return Vec<decltype(a.x op b.y), 4>{(a.x op b.x), (a.y op b.y), (a.z op b.z), (a.w op b.w)};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 4> operator op(const Vec<T1, 4> &a, const     T2     &b) {return Vec<decltype(a.x op b  ), 4>{(a.x op b  ), (a.y op b  ), (a.z op b  ), (a.w op b  )};}\
		                          template <typename T1, typename T2> Vec<decltype(VAL(T1) op VAL(T2)), 4> operator op(const     T1     &a, const Vec<T2, 4> &b) {return Vec<decltype(a   op b.y), 4>{(a   op b.x), (a   op b.y), (a   op b.z), (a   op b.w)};}
		#define OP1_2A(op)        template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 1> &operator op(Vec<T1, 1> &a, const Vec<T2, 1> &b) {a.x op b.x; return a;}\
		                          template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 1> &operator op(Vec<T1, 1> &a, const     T2     &b) {a.x op b  ; return a;}
		#define OP2_2A(op)        template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 2> &operator op(Vec<T1, 2> &a, const Vec<T2, 2> &b) {a.x op b.x; a.y op b.y; return a;}\
		                          template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 2> &operator op(Vec<T1, 2> &a, const     T2     &b) {a.x op b  ; a.y op b  ; return a;}
		#define OP3_2A(op)        template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 3> &operator op(Vec<T1, 3> &a, const Vec<T2, 3> &b) {a.x op b.x; a.y op b.y; a.z op b.z; return a;}\
		                          template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 3> &operator op(Vec<T1, 3> &a, const     T2     &b) {a.x op b  ; a.y op b  ; a.z op b  ; return a;}
		#define OP4_2A(op)        template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 4> &operator op(Vec<T1, 4> &a, const Vec<T2, 4> &b) {a.x op b.x; a.y op b.y; a.z op b.z; a.w op b.w; return a;}\
		                          template <typename T1, typename T2> Vec<typename std::remove_reference<decltype(VAL(T1) op VAL(T2))>::type, 4> &operator op(Vec<T1, 4> &a, const     T2     &b) {a.x op b  ; a.y op b  ; a.z op b  ; a.w op b  ; return a;}

		#define OP_PACK(X)    \
		OP##X##_1(+)          \
		OP##X##_1(-)          \
		OP##X##_1(!)          \
		OP##X##_1(~)          \
		OP##X##_1(++)         \
		OP##X##_1(--)         \
		OP##X##_1_POSTFIX(++) \
		OP##X##_1_POSTFIX(--) \
		OP##X##_2(+)          \
		OP##X##_2(-)          \
		OP##X##_2(*)          \
		OP##X##_2(/)          \
		OP##X##_2(%)          \
		OP##X##_2(^)          \
		OP##X##_2(&)          \
		OP##X##_2(|)          \
		OP##X##_2(&&)         \
		OP##X##_2(||)         \
		OP##X##_2(>)          \
		OP##X##_2(<)          \
		OP##X##_2(>=)         \
		OP##X##_2(<=)         \
		OP##X##_2(<<)         \
		OP##X##_2(>>)         \
		OP##X##_2A(+=)        \
		OP##X##_2A(-=)        \
		OP##X##_2A(*=)        \
		OP##X##_2A(/=)        \
		OP##X##_2A(%=)        \
		OP##X##_2A(^=)        \
		OP##X##_2A(&=)        \
		OP##X##_2A(|=)        \
		OP##X##_2A(<<=)       \
		OP##X##_2A(>>=)

		template <typename T, unsigned int D>
		struct Vec;

		template <typename T>
		struct Vec<T, 1>
		{
			COMMON
			static constexpr int dim = 1;
			union {T x, r;};

			operator bool() {return bool(x);}
			template <typename TT> operator Vec<TT, 1>() const {return {TT(x)};}
		};
		OP_PACK(1)

		template <typename T>
		struct Vec<T, 2>
		{
			COMMON
			static constexpr int dim = 2;
			union {T x, r;};
			union {T y, g;};

			operator bool() {return bool(x) || bool(y);}
			operator Vec<T, 1>() const {return {x};}
			template <typename TT> operator Vec<TT, 2>() const {return {TT(x), TT(y)};}
		};
		OP_PACK(2)

		template <typename T>
		struct Vec<T, 3>
		{
			COMMON
			static constexpr int dim = 3;
			union {T x, r;};
			union {T y, g;};
			union {T z, b;};

			operator bool() {return bool(x) || bool(y) || bool(z);}
			operator Vec<T, 1>() const {return {x};}
			operator Vec<T, 2>() const {return {x, y};}
			template <typename TT> operator Vec<TT, 3>() const {return {TT(x), TT(y), TT(z)};}
		};
		OP_PACK(3)

		template <typename T>
		struct Vec<T, 4>
		{
			COMMON
			static constexpr int dim = 4;
			union {T x, r;};
			union {T y, g;};
			union {T z, b;};
			union {T w, a;};

			operator bool() {return bool(x) || bool(y) || bool(z) || bool(w);}
			operator Vec<T, 1>() const {return {x};}
			operator Vec<T, 2>() const {return {x, y};}
			operator Vec<T, 3>() const {return {x, y, z};}
			template <typename TT> operator Vec<TT, 4>() const {return {TT(x), TT(y), TT(z), TT(w)};}
		};
		OP_PACK(4)

		template <typename T1, typename T2> bool operator==(const Vec<T1, 1> &a, const Vec<T2, 1> &b) {return a.x == b.x;}
		template <typename T1, typename T2> bool operator==(const Vec<T1, 2> &a, const Vec<T2, 2> &b) {return a.x == b.x && a.y == b.y;}
		template <typename T1, typename T2> bool operator==(const Vec<T1, 3> &a, const Vec<T2, 3> &b) {return a.x == b.x && a.y == b.y && a.z == b.z;}
		template <typename T1, typename T2> bool operator==(const Vec<T1, 4> &a, const Vec<T2, 4> &b) {return a.x == b.x && a.y == b.y && a.z == b.z && a.w == b.w;}
		template <typename T1, typename T2, unsigned int D> bool operator!=(const Vec<T1, D> &a, const Vec<T2, D> &b) {return!(a == b);}
		template <typename T1, typename T2> bool operator==(const Vec<T1, 1> &a, const     T2     &b) {return a.x == b  ;}
		template <typename T1, typename T2> bool operator==(const Vec<T1, 2> &a, const     T2     &b) {return a.x == b   && a.y == b  ;}
		template <typename T1, typename T2> bool operator==(const Vec<T1, 3> &a, const     T2     &b) {return a.x == b   && a.y == b   && a.z == b  ;}
		template <typename T1, typename T2> bool operator==(const Vec<T1, 4> &a, const     T2     &b) {return a.x == b   && a.y == b   && a.z == b   && a.w == b  ;}
		template <typename T1, typename T2, unsigned int D> bool operator!=(const Vec<T1, D> &a, const     T2     &b) {return!(a == b);}
		template <typename T1, typename T2> bool operator==(const     T1     &a, const Vec<T2, 1> &b) {return a   == b.x;}
		template <typename T1, typename T2> bool operator==(const     T1     &a, const Vec<T2, 2> &b) {return a   == b.x && a   == b.y;}
		template <typename T1, typename T2> bool operator==(const     T1     &a, const Vec<T2, 3> &b) {return a   == b.x && a   == b.y && a   == b.z;}
		template <typename T1, typename T2> bool operator==(const     T1     &a, const Vec<T2, 4> &b) {return a   == b.x && a   == b.y && a   == b.z && a   == b.w;}
		template <typename T1, typename T2, unsigned int D> bool operator!=(const     T1     &a, const Vec<T2, D> &b) {return!(a == b);}

		#undef COMMON
		#undef OP1_1
		#undef OP2_1
		#undef OP3_1
		#undef OP4_1
		#undef OP1_1_POSTFIX
		#undef OP2_1_POSTFIX
		#undef OP3_1_POSTFIX
		#undef OP4_1_POSTFIX
		#undef OP1_2
		#undef OP2_2
		#undef OP3_2
		#undef OP4_2
		#undef OP1_2A
		#undef OP2_2A
		#undef OP3_2A
		#undef OP4_2A
		#undef OP_PACK

		template <typename T = float> using vec1 = Vec<T, 1>;
		template <typename T = float> using vec2 = Vec<T, 2>;
		template <typename T = float> using vec3 = Vec<T, 3>;
		template <typename T = float> using vec4 = Vec<T, 4>;

		#define USING(a, b) using a##vec1 = Vec<b, 1>; using a##vec2 = Vec<b, 2>; using a##vec3 = Vec<b, 3>; using a##vec4 = Vec<b, 4>;
		USING(f,  float)
		USING(d,  double)
		USING(ld, long double)
		USING(c,  signed char)
		USING(uc, unsigned char)
		USING(s,  short)
		USING(us, unsigned short)
		USING(i,  int)
		USING(ui, unsigned int)
		USING(l,  long)
		USING(ul, unsigned long)
		USING(ll, long long)
		USING(ull,unsigned long long)
		#undef USING

		template <typename T, unsigned int D> Vec<T, D> Len2(const Vec<T, D> &in);
		template <typename T> T Len2(const Vec<T, 1> &in) {return in.x*in.x;}
		template <typename T> T Len2(const Vec<T, 2> &in) {return in.x*in.x + in.y*in.y;}
		template <typename T> T Len2(const Vec<T, 3> &in) {return in.x*in.x + in.y*in.y + in.z*in.z;}
		template <typename T> T Len2(const Vec<T, 4> &in) {return in.x*in.x + in.y*in.y + in.z*in.z + in.w*in.w;}
		template <typename T, unsigned int D> FP_Type<T> Len(const Vec<T, D> &in) {FP_Type<T> ret = Len2<T>(in); return ret ? std::sqrt(ret) : 0;}
		template <typename T> FP_Type<T> Len(const Vec<T, 1> &in) {return std::abs(in.x);}

		template <typename T, unsigned int D> Vec<T, D> Normalize(const Vec<T, D> &in)
		{
			FP_Type<T> len = Len<T, D>(in);
			if (len == 0)
				return Vec<T, D>();
			else
				return in / len;
		}
		template <typename T> Vec<T, 1> Normalize(const Vec<T, 1> &in)
		{
			return {in.x ? (in.x > 0 ? 1 : -1) : 0};
		}

		template <typename T1, typename T2> decltype(VAL(T1) * VAL(T2) + VAL(T1) * VAL(T2)) Dot(const Vec<T1, 1> &a, const Vec<T2, 1> &b) {return a.x * b.x;}
		template <typename T1, typename T2> decltype(VAL(T1) * VAL(T2) + VAL(T1) * VAL(T2)) Dot(const Vec<T1, 2> &a, const Vec<T2, 2> &b) {return a.x * b.x + a.y * b.y;}
		template <typename T1, typename T2> decltype(VAL(T1) * VAL(T2) + VAL(T1) * VAL(T2)) Dot(const Vec<T1, 3> &a, const Vec<T2, 3> &b) {return a.x * b.x + a.y * b.y + a.z * b.z;}
		template <typename T1, typename T2> decltype(VAL(T1) * VAL(T2) + VAL(T1) * VAL(T2)) Dot(const Vec<T1, 4> &a, const Vec<T2, 4> &b) {return a.x * b.x + a.y * b.y + a.z * b.z + a.w * b.w;}

		template <typename T1, typename T2> decltype(VAL(T1) * VAL(T2) - VAL(T1) * VAL(T2)) Cross(const Vec<T1, 3> &a, const Vec<T2, 3> &b)
		{
			return Vec<decltype(VAL(T1) * VAL(T2) - VAL(T1) * VAL(T2)), 3>{a.y * b.z - a.z * b.y,
																		   a.z * b.x - a.x * b.z,
																		   a.x * b.y - a.y * b.x};
		}

		#undef VAL

		// END OF VEC DEFINITION

		class Quaternion
		{
		  public:
			float x, y, z, w;

			Quaternion()
			{
				x = y = z = 0;
				w = 1;
			}
			Quaternion(Vec<float, 3> v, float angle)
			{
				float d = std::sqrt(v.x*v.x + v.y*v.y + v.z*v.z);
				if (d == 0)
				{
					Quaternion();
					return;
				}
				v.x /= d;
				v.y /= d;
				v.z /= d;
				x = v.x * std::sin(angle / 2.f);
				y = v.y * std::sin(angle / 2.f);
				z = v.z * std::sin(angle / 2.f);
				w = std::cos(angle / 2.f);
			}
			Quaternion(Vec<float, 3> v1, Vec<float, 3> v2) // vec to vec rotation
			{
				float newx =  v1.y*v2.z - v1.z*v2.y;
				float newy =  v1.z*v2.x - v1.x*v2.z;
				float newz =  v1.x*v2.y - v1.y*v2.x;
				if (std::sqrt(newx*newx + newy*newy + newz*newz) == 0.0)
					Quaternion();
				else
				{
					float cos = (v1.x*v2.x + v1.y*v2.y + v1.z*v2.z) / std::sqrt(v1.x*v1.x + v1.y*v1.y + v1.z*v1.z) / std::sqrt(v2.x*v2.x + v2.y*v2.y + v2.z*v2.z);
					Quaternion({newx, newy, newz}, std::acos(cos * fPi / 180));
				}
			}
			void Normalize()
			{
				float d = std::sqrt(x*x + y*y + z*z + w*w);
				x /= d;
				y /= d;
				z /= d;
				w /= d;
			}
			Quaternion operator+(const Quaternion &other) const
			{
				float newx, newy, newz, neww, d;
				newx = x + other.x;
				newy = y + other.y;
				newz = z + other.z;
				neww = w + other.w;
				d = std::sqrt(newx*newx + newy*newy + newz*newz + neww*neww);
				return new_q(newx / d, newy / d, newz / d, neww / d);
			}
			Quaternion operator-(const Quaternion &other) const
			{
				float newx, newy, newz, neww, d;
				newx = x - other.x;
				newy = y - other.y;
				newz = z - other.z;
				neww = w - other.w;
				d = std::sqrt(newx*newx + newy*newy + newz*newz + neww*neww);
				return new_q(newx / d, newy / d, newz / d, neww / d);
			}
			Quaternion operator-() const
			{
				return new_q(-x, -y, -z, w);
			}
			Quaternion operator*(const Quaternion &other) const
			{
				//if (other.GetAngle() == 0.0) return *this;
				float newx, newy, newz, neww, d;

				newx = w*other.x + x*other.w + y*other.z - z*other.y;
				newy = w*other.y - x*other.z + y*other.w + z*other.x;
				newz = w*other.z + x*other.y - y*other.x + z*other.w;
				neww = w*other.w - x*other.x - y*other.y - z*other.z;

				d = std::sqrt(newx*newx + newy*newy + newz*newz + neww*neww);
				return new_q(newx / d, newy / d, newz / d, neww / d);
			}
			Quaternion &operator*=(const Quaternion &other)
			{
				//if (other.GetAngle() == 0.0) return *this;
				float newx, newy, newz, neww, d;

				newx = w*other.x + x*other.w + y*other.z - z*other.y;
				newy = w*other.y - x*other.z + y*other.w + z*other.x;
				newz = w*other.z + x*other.y - y*other.x + z*other.w;
				neww = w*other.w - x*other.x - y*other.y - z*other.z;

				d = std::sqrt(newx*newx + newy*newy + newz*newz + neww*neww);
				x = newx / d;
				y = newy / d;
				z = newz / d;
				w = neww / d;
				return *this;
			}
			float operator%(Quaternion other) const // dot product
			{
				return x*other.x + y*other.y + z*other.z + w*other.w;
			}
			float LenSquared() const
			{
				return x*x + y*y + z*z + w*w;
			}
			float Len() const
			{
				return std::sqrt(LenSquared());
			}
			Quaternion Combine(const Quaternion &other, float t1, float t2) const
			{
				float newx, newy, newz, neww, d;
				newx = x*t1 + other.x*t2;
				newy = y*t1 + other.y*t2;
				newz = z*t1 + other.z*t2;
				neww = w*t1 + other.w*t2;
				d = std::sqrt(newx*newx + newy*newy + newz*newz + neww*neww);
				return new_q(newx / d, newy / d, newz / d, neww / d);
			}

			bool operator==(const Quaternion &other) const
			{
				return x == other.x && y == other.y && z == other.z && w == other.w;
			}
			bool operator!=(const Quaternion &other) const
			{
				return !(x == other.x && y == other.y && z == other.z && w == other.w);
			}

			Vec<float, 3> GetAxis() const
			{
				float d = std::sqrt(x*x + y*y + z*z);
				return d ? Vec<float, 3>{x/d, y/d, z/d} : Vec<float, 3>{0, 0, 0};
			}
			float GetAngle() const
			{
				return std::atan2(std::sqrt(x*x + y*y + z*z), w) * 2;
			}

			Quaternion VarRotate(float n) const
			{
				return Quaternion({x, y, z}, GetAngle()*n);
			}

			Vec<float, 3> SlowTransform(const Vec<float, 3> &in) const // 24x [*|/]  17x [+|-]
			{
				float newx = w*in.x + y*in.z - z*in.y;
				float newy = w*in.y - x*in.z + z*in.x;
				float newz = w*in.z + x*in.y - y*in.x;
				float neww = x*in.x + y*in.y + z*in.z;

				return {newx*w + neww*x - newy*z + newz*y,
						neww*y + newx*z + newy*w - newz*x,
						neww*z - newx*y + newy*x + newz*w};
			}

			void PrepareMassTransform() const // 18x [*|/]  12 [+|-]
			{
				t[0][0] = 1 - 2*y*y - 2*z*z; t[1][0] = 2*x*y - 2*z*w; t[2][0] = 2*x*z + 2*y*w;
				t[0][1] = 2*x*y + 2*z*w; t[1][1] = 1 - 2*x*x - 2*z*z; t[2][1] = 2*y*z - 2*x*w;
				t[0][2] = 2*x*z - 2*y*w; t[1][2] = 2*y*z + 2*x*w; t[2][2] = 1 - 2*x*x - 2*y*y;
			}

			Vec<float, 3> MassTransform(const Vec<float, 3> &in) const // 9x [*|/]  6x [+|-] // Uses last quaternion for which PrepareMassTransform() was called
			{
				return {in.x*t[2][0] + in.y*t[2][0] + in.z*t[2][0],
						in.x*t[2][1] + in.y*t[2][1] + in.z*t[2][1],
						in.x*t[2][2] + in.y*t[2][2] + in.z*t[2][2]};
			}

			static inline float (*GetMassTransformMatrix3x3())[3] // Returns pointer to static storage, matrix changes after each PrepareMassTransform() call.
			{
				return t;
			}

			static inline Quaternion new_q(float a, float b, float c, float d)
			{
				Quaternion ret;
				ret.x = a;
				ret.y = b;
				ret.z = c;
				ret.w = d;
				return ret;
			}

		  private:
			static float t[3][3];
		};
	}

	// * TypeAt
	// TypeAt<unsigned, Type1, Type2, ...>::type
	// Example: TypeAt<1, short, int, long>::type == int
	//
	// * IsTypeInList
	// IsTypeInList<BaseType, Type1, Type2, ...>::value
	// Example: IsTypeInList<int, char, short, long>::value      == 0
	//          IsTypeInList<int, char, short, int, long>::value == 1
	// Also works with 0 additional types
	namespace Template
	{
		template <unsigned int X, typename T, typename ...P> struct TypeAt {using type = TypeAt<X-1, P...>;};
		template <typename T, typename ...P> struct TypeAt<0, T, P...> {using type = T;};

		namespace Internal
		{
			template <typename T, typename F, typename ...P> struct IsTypeInList
			{
				static constexpr bool value = std::is_same<T, F>::value || IsTypeInList<T, P...>::value;
			};
			template <typename T, typename F> struct IsTypeInList<T, F>
			{
				static constexpr bool value = std::is_same<T, F>::value;
			};
		}

		template <typename T, typename ...P> struct IsTypeInList
		{
			static constexpr bool value = Internal::IsTypeInList<T, P...>::value;
		};
		template <typename T> struct IsTypeInList<T>
		{
			static constexpr bool value = 0;
		};

	}

	inline namespace Containers
	{
		template <unsigned int Size, unsigned int Align> using Storage = typename std::aligned_storage<Size, Align>::type;
		template <typename T> using TypeStorage = typename std::aligned_storage<sizeof (T), alignof (T)>::type;

		inline namespace ObjectBox
		{

			namespace Internal
			{
				#define PACK_PARAMETER P ... p
				#define PACK_EXPANSION std::forward<P>(p)...

				#define NORMAL_CTORS \
					template <typename ...P> void Ctor(PACK_PARAMETER)\
					{\
						MemModule<T, Heap>::Mem_AtCtor();\
						new(MemModule<T, Heap>::Mem_Ptr()) T(PACK_EXPANSION);\
					}\
					template <typename ...P> void BracedCtor(PACK_PARAMETER)\
					{\
						MemModule<T, Heap>::Mem_AtCtor();\
						new(MemModule<T, Heap>::Mem_Ptr()) T{PACK_EXPANSION};\
					}

				#define NORMAL_DTOR \
					void Dtor()\
					{\
						MemModule<T, Heap>::Mem_Ptr()->~T();\
						MemModule<T, Heap>::Mem_AtDtor();\
					}

				#define NORMAL_RECREATE \
					template <typename ...P> void Recreate(PACK_PARAMETER)\
					{\
						MemModule<T, Heap>::Mem_Ptr()->~T();\
						new(MemModule<T, Heap>::Mem_Ptr()) T(PACK_EXPANSION);\
					}

				#define ARRAY_CTORS \
					template <typename ...P> void Ctor(PACK_PARAMETER)\
					{\
						MemModule<T, Heap>::Mem_AtCtor();\
						new(MemModule<T, Heap>::Mem_Ptr()) T{PACK_EXPANSION};\
					}

				#define ARRAY_DTOR \
					void Dtor()\
					{\
						for (Type &it : *MemModule<T, Heap>::Mem_Ptr())\
							it.~Type();\
						MemModule<T, Heap>::Mem_AtDtor();\
					}

				#define ARRAY_RECREATE \
					template <typename ...P> void Recreate(PACK_PARAMETER)\
					{\
						for (Type &it : *MemModule<T, Heap>::Mem_Ptr())\
							it.~Type();\
						new(MemModule<T, Heap>::Mem_Ptr()) T{PACK_EXPANSION};\
					}



				template <typename T, unsigned int Heap> class MemModule;
				template <typename T> class MemModule<T, 0> // No heap
				{
					TypeStorage<T> data;
				  protected:
				  	void Mem_AtBoxCtor() {}
				  	void Mem_AtBoxDtor() {}
					void Mem_AtCtor()    {}
					void Mem_AtDtor()    {}
					T *Mem_Ptr()         {return (T *)&data;}
					const T *Mem_Ptr() const {return (T *)&data;}
				};
				template <typename T> class MemModule<T, 1> // Box livetime heap storage
				{
					T *heap;
				  protected:
					void Mem_AtBoxCtor() {heap = (T *)new TypeStorage<T>;}
				  	void Mem_AtBoxDtor() {delete (TypeStorage<T> *)heap;}
					void Mem_AtCtor()    {}
					void Mem_AtDtor()    {}
					T *Mem_Ptr()         {return heap;}
					const T *Mem_Ptr() const {return heap;}
				};
				template <typename T> class MemModule<T, 2> // Obj livetime heap storage
				{
					T *heap;
				  protected:
				  	void Mem_AtBoxCtor() {heap = 0;}
				  	void Mem_AtBoxDtor() {if (heap) delete (TypeStorage<T> *)heap;}
					void Mem_AtCtor()    {heap = (T *)new TypeStorage<T>;}
					void Mem_AtDtor()    {delete (TypeStorage<T> *)heap;}
					T *Mem_Ptr()         {return heap;}
					const T *Mem_Ptr() const {return heap;}
				};


				template <typename T, unsigned int Heap, bool ManCtor, bool ManDtor> class MemLifetimeModule;
				template <typename T, unsigned int Heap>
				class MemLifetimeModule<T, Heap, 0, 0> : public MemModule<T, Heap>
				{
				  protected:
				  	using Base = MemModule<T, Heap>;
					NORMAL_CTORS
					NORMAL_DTOR
				  public:
					NORMAL_RECREATE
				};
				template <typename T, unsigned int Heap>
				class MemLifetimeModule<T, Heap, 0, 1> : public MemModule<T, Heap>
				{
				  protected:
				  	using Base = MemModule<T, Heap>;
					NORMAL_CTORS
				  public:
				  	NORMAL_DTOR
				  	NORMAL_RECREATE
				};
				template <typename T, unsigned int Heap>
				class MemLifetimeModule<T, Heap, 1, 0> : public MemModule<T, Heap>
				{
				  public:
					NORMAL_CTORS
					NORMAL_RECREATE
				  protected:
				  	NORMAL_DTOR
				  	using Base = MemModule<T, Heap>;
				};
				template <typename T, unsigned int Heap>
				class MemLifetimeModule<T, Heap, 1, 1> : public MemModule<T, Heap>
				{
				  protected:
					using Base = MemModule<T, Heap>;
				  public:
					NORMAL_CTORS
					NORMAL_DTOR
					NORMAL_RECREATE
				};


				template <typename Type, std::size_t X, unsigned int Heap>
				class MemLifetimeModule<Type[X], Heap, 0, 0> : public MemModule<Type[X], Heap>
				{
					using T = Type[X];
				  protected:
				  	using Base = MemModule<T, Heap>;
					ARRAY_CTORS
					ARRAY_DTOR
				  public:
					ARRAY_RECREATE
				};
				template <typename Type, std::size_t X, unsigned int Heap>
				class MemLifetimeModule<Type[X], Heap, 0, 1> : public MemModule<Type[X], Heap>
				{
					using T = Type[X];
				  protected:
				  	using Base = MemModule<T, Heap>;
					ARRAY_CTORS
				  public:
				  	ARRAY_DTOR
				  	ARRAY_RECREATE
				};
				template <typename Type, std::size_t X, unsigned int Heap>
				class MemLifetimeModule<Type[X], Heap, 1, 0> : public MemModule<Type[X], Heap>
				{
					using T = Type[X];
				  public:
					ARRAY_CTORS
					ARRAY_RECREATE
				  protected:
				  	ARRAY_DTOR
				  	using Base = MemModule<T, Heap>;
				};
				template <typename Type, std::size_t X, unsigned int Heap>
				class MemLifetimeModule<Type[X], Heap, 1, 1> : public MemModule<Type[X], Heap>
				{
					using T = Type[X];
				  protected:
					using Base = MemModule<T, Heap>;
				  public:
					ARRAY_CTORS
					ARRAY_DTOR
					ARRAY_RECREATE
				};

				#undef NORMAL_CTORS
				#undef NORMAL_DTOR
				#undef NORMAL_RECREATE
				#undef ARRAY_CTORS
				#undef ARRAY_DTOR
				#undef ARRAY_RECREATE

				template <typename Type, unsigned int Heap, bool ManCtor, bool ManDtor>
				class ExtensionModule : public MemLifetimeModule<Type, Heap, ManCtor, ManDtor>
				{
					using T = Type;
				  protected:
					using Base = MemLifetimeModule<T, Heap, ManCtor, ManDtor>;
				  public:
					T &operator*() {return *MemModule<T, Heap>::Mem_Ptr();}
					const T &operator*() const {return *MemModule<T, Heap>::Mem_Ptr();}
					T *operator->() {return MemModule<T, Heap>::Mem_Ptr();}
					const T *operator->() const {return MemModule<T, Heap>::Mem_Ptr();}
					operator T *() {return MemModule<T, Heap>::Mem_Ptr();}
					operator const T *() const {return MemModule<T, Heap>::Mem_Ptr();}
				};

				template <typename Type, typename ...P, unsigned int Heap, bool ManCtor, bool ManDtor>
				class ExtensionModule<Type (*)(P...), Heap, ManCtor, ManDtor> : public MemLifetimeModule<Type (*)(P...), Heap, ManCtor, ManDtor>
				{
				  protected:
				  	using T = Type (*)(P...);
					using Base = MemLifetimeModule<T, Heap, ManCtor, ManDtor>;
				  public:
					operator T *() {return MemModule<T, Heap>::Mem_Ptr();}
					operator const T *() const {return MemModule<T, Heap>::Mem_Ptr();}
					Type operator()(P &&... p)
					{
						return std::forward<Type>((*MemModule<T, Heap>::Mem_Ptr())(std::forward<P>(p)...));
					}
				};

				template <typename Type, std::size_t X, unsigned int Heap, bool ManCtor, bool ManDtor>
				class ExtensionModule<Type[X], Heap, ManCtor, ManDtor> : public MemLifetimeModule<Type[X], Heap, ManCtor, ManDtor>
				{
				  protected:
				  	using T = Type[X];
					using Base = MemLifetimeModule<T, Heap, ManCtor, ManDtor>;
				  public:
					operator Type *() {return (Type *)MemModule<T, Heap>::Mem_Ptr();}
					operator const Type *() const {return (const Type *)MemModule<T, Heap>::Mem_Ptr();}
					Type &operator*() {return *(Type *)MemModule<T, Heap>::Mem_Ptr();}
					const Type &operator*() const {return *(Type *)MemModule<T, Heap>::Mem_Ptr();}
				};

				template <typename T, unsigned int Heap, bool ManCtor, bool ManDtor, bool NoCtor, bool NoDtor> class Box;
				template <typename T, unsigned int Heap, bool ManCtor, bool ManDtor, bool NoDtor>
				class Box<T, Heap, ManCtor, ManDtor, 0, NoDtor> : public ExtensionModule<T, Heap, ManCtor, ManDtor>
				{
					using Base = ExtensionModule<T, Heap, ManCtor, ManDtor>;
				  public:
					template <typename ... P> Box(P &&... p)
					{
						Base::Base::Base::Mem_AtBoxCtor();
						Base::Base::Ctor(std::forward<decltype(p)>(p)...);
					}
					~Box()
					{
						if (!NoDtor)
							Base::Base::Dtor();
						Base::Base::Base::Mem_AtBoxDtor();
					}
				};
				template <typename T, unsigned int Heap, bool ManCtor, bool ManDtor, bool NoDtor>
				class Box<T, Heap, ManCtor, ManDtor, 1, NoDtor> : public ExtensionModule<T, Heap, ManCtor, ManDtor>
				{
					using Base = ExtensionModule<T, Heap, ManCtor, ManDtor>;
				  public:
					Box()
					{
						MemModule<T, Heap>::Mem_AtBoxCtor();
					}
					~Box()
					{
						if (!NoDtor)
							MemLifetimeModule<T, Heap, ManCtor, ManDtor>::Dtor();
						MemModule<T, Heap>::Mem_AtBoxDtor();
					}
				};

				template <typename T> struct TypeFixup
				{
					using type = T;
				};
				template <typename T, typename ...P> struct TypeFixup<T (P...)>
				{
					using type = T (*)(P...);
				};

				template <typename Type, typename ...Cfg> struct CfgBox
				{
					#define Option(x) (Template::IsTypeInList<typename Tags::x, Cfg...>::value)
					static_assert(!(Option(Heap) && Option(Heap_DelUnused)), "Conflicting flags.");
					using type = Internal::Box<typename TypeFixup<Type>::type, (Option(Heap)               ? 1 : 0) +
																			   (Option(Heap_DelUnused)     ? 2 : 0),
																			   Option(ManLifetime) ||       //
																			   Option(NoAutoCtor) ||        //
																			   Option(NoAutoCtorDtor),      //
																			   Option(ManLifetime) ||         //
																			   Option(NoAutoDtor) ||          //
																			   Option(NoAutoCtorDtor),        //
																			   Option(NoAutoCtor) ||        //
																			   Option(NoAutoCtorDtor),      //
																			   Option(NoAutoDtor) ||          //
																			   Option(NoAutoCtorDtor)>;       //
					#undef Option
				};
			}

			template <typename Type, typename ...Cfg>
			class Box : public Internal::CfgBox<Type, Cfg...>::type
			{
			  public:
				template <typename ... P> Box(P &&... p) : Internal::CfgBox<Type, Cfg...>::type::Box(p...) {}
				Box(const Box &) = delete;
				Box(Box &&) = delete;
				Box &operator=(const Box &) = delete;
				Box &operator=(Box &&) = delete;
			};
		}
	}

	inline uint64_t Clock()               {return SDL_GetPerformanceCounter();}
	inline uint64_t ClockTicksPerSecond() {static uint64_t ret = SDL_GetPerformanceFrequency(); return ret;}
	inline uint64_t SecondsToClockUnits(long double secs) {return secs * ClockTicksPerSecond();}
	inline uint64_t ClockUnitsToSecs(uint64_t units)   {return (long double)(units / ClockTicksPerSecond()) + (long double)(units % ClockTicksPerSecond()) / (long double)ClockTicksPerSecond();}

	inline void WaitSecs(long double secs)
	{
		uint64_t delay = secs * (long double)(ClockTicksPerSecond());
		uint64_t begin = Clock();
		while (Clock() - begin < delay)
		{
		#if MANIFEST_FEATURE_CRAPPY_TIMER == 1
		SDL_Delay(1);
		#endif
		}
	}
	inline void WaitClockUnits(uint64_t delay)
	{
		uint64_t begin = Clock();
		while (Clock() - begin < delay)
		{
		#if MANIFEST_FEATURE_CRAPPY_TIMER == 1
		SDL_Delay(1);
		#endif
		}
	}
	inline bool CheckIfFileExists(const char *name)
	{
		FILE *f = std::fopen(name, "r");
		if (f)
		{
			std::fclose(f);
			return 1;
		}
		else
			return 0;
	}

	inline namespace StringJoiner
	{
		class Internal final
		{
			~Internal() = delete;
			static std::stringstream stream;
			static const std::ios_base::fmtflags defflags;
			static std::string ret;
			template <typename T, typename ...P> struct Append
			{
				static void func(const void *const *p)
				{
					stream << *(typename std::remove_reference<T>::type *)*p;
					Append<P...>::func(p + 1);
				}
			};
			template <typename T> struct Append<T>
			{
				static void func(const void *const *p)
				{
					stream << *(typename std::remove_reference<T>::type *)*p;
				}
			};
			template <typename ...P> friend const char *Jo(P &&... p);
		};

		template <typename ...P> const char *Jo(P &&... p) // 'Join', returned pointer is valid until next call
		{
			Internal::stream.clear();
			Internal::stream.flags(Internal::defflags);
			Internal::stream.str("");
			const void *const arr[sizeof...(P)] {&p...};
			Internal::Append<P...>::func(arr);
			Internal::ret = Internal::stream.str();
			return Internal::ret.c_str();
		}
	}


	static constexpr bool is_char_signed = char(-1) < 0;

	template <int X, int Y, typename T = float> inline T ScalarFromMat4Product(const T (*a)[4], const T (*b)[4]) // Pro tip: only b[X][...] is used
	{
		static_assert(X >= 0 && Y >= 0 && X < 4 && Y < 4, "Index is out of range.");
		return a[0][Y]*b[X][0] + a[1][Y]*b[X][1] + a[2][Y]*b[X][2] + a[3][Y]*b[X][3];
	}
}

//using Utils::Str;
using Utils::Jo;
using Utils::DivCeil;
namespace M = Utils::Math;

#endif